@@ -32,8 +32,26 @@
 #define SELECTION_COUNT (VD_AGENT_CLIPBOARD_SELECTION_PRIMARY + 1)
 #define TYPE_COUNT      (VD_AGENT_CLIPBOARD_IMAGE_JPG + 1)
 
-#define sel_id_from_clip(clipboard) \
-    GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(clipboard), "vdagent-selection-id"))
+static const GdkAtom sel_atom[] = {
+    [VD_AGENT_CLIPBOARD_SELECTION_CLIPBOARD] = GDK_SELECTION_CLIPBOARD,
+    [VD_AGENT_CLIPBOARD_SELECTION_PRIMARY] = GDK_SELECTION_PRIMARY,
+};
+
+G_STATIC_ASSERT(G_N_ELEMENTS(sel_atom) == SELECTION_COUNT);
+
+static gint sel_id_from_clip(GtkClipboard *clipboard)
+{
+    GdkAtom sel = gtk_clipboard_get_selection(clipboard);
+    int i;
+
+    for (i = 0; i < G_N_ELEMENTS(sel_atom); i++) {
+        if (sel == sel_atom[i]) {
+            return i;
+        }
+    }
+
+    g_return_val_if_reached(0);
+}
 
 enum {
     OWNER_NONE,
@@ -58,15 +76,25 @@ typedef struct {
 } Selection;
 #endif
 
-struct VDAgentClipboards {
+struct _VDAgentClipboards {
+    GObject parent;
+
+    UdscsConnection *conn;
+
 #ifdef WITH_GTK
-    struct udscs_connection *conn;
-    Selection                selections[SELECTION_COUNT];
+    Selection selections[SELECTION_COUNT];
 #else
     struct vdagent_x11 *x11;
 #endif
 };
 
+struct _VDAgentClipboardsClass
+{
+    GObjectClass parent;
+};
+
+G_DEFINE_TYPE(VDAgentClipboards, vdagent_clipboards, G_TYPE_OBJECT)
+
 #ifdef WITH_GTK
 static const struct {
     guint         type;
@@ -210,16 +238,17 @@ static void clipboard_owner_change_cb(GtkClipboard        *clipboard,
     Selection *sel = &c->selections[sel_id];
 
     /* if the event was caused by gtk_clipboard_set_with_data(), ignore it  */
-    if (sel->owner == OWNER_CLIENT)
+    if (gtk_clipboard_get_owner(clipboard) == G_OBJECT(c)) {
         return;
-
-    if (sel->owner == OWNER_GUEST) {
-        clipboard_new_owner(c, sel_id, OWNER_NONE);
-        udscs_write(c->conn, VDAGENTD_CLIPBOARD_RELEASE, sel_id, 0, NULL, 0);
     }
 
-    if (event->reason != GDK_OWNER_CHANGE_NEW_OWNER)
+    if (event->reason != GDK_OWNER_CHANGE_NEW_OWNER) {
+        if (sel->owner == OWNER_GUEST) {
+            clipboard_new_owner(c, sel_id, OWNER_NONE);
+            udscs_write(c->conn, VDAGENTD_CLIPBOARD_RELEASE, sel_id, 0, NULL, 0);
+        }
         return;
+    }
 
     /* if there's a pending request for clipboard targets, cancel it */
     if (sel->last_targets_req)
@@ -329,9 +358,10 @@ void vdagent_clipboard_grab(VDAgentClipboards *c, guint sel_id,
         g_clear_pointer(&sel->last_targets_req, request_ref_cancel);
     }
 
-    if (gtk_clipboard_set_with_data(sel->clipboard,
-                                    targets, n_targets,
-                                    clipboard_get_cb, clipboard_clear_cb, c))
+    if (gtk_clipboard_set_with_owner(sel->clipboard,
+                                     targets, n_targets,
+                                     clipboard_get_cb, clipboard_clear_cb,
+                                     G_OBJECT(c)))
         clipboard_new_owner(c, sel_id, OWNER_CLIENT);
     else {
         syslog(LOG_ERR, "%s: sel_id=%u: clipboard grab failed", __func__, sel_id);
@@ -435,50 +465,56 @@ err:
 #endif
 }
 
-VDAgentClipboards *vdagent_clipboards_init(struct vdagent_x11      *x11,
-                                           struct udscs_connection *conn)
+static void
+vdagent_clipboards_init(VDAgentClipboards *self)
 {
-#ifdef WITH_GTK
-    guint sel_id;
-    const GdkAtom sel_atom[SELECTION_COUNT] = {
-        GDK_SELECTION_CLIPBOARD, /* VD_AGENT_CLIPBOARD_SELECTION_CLIPBOARD */
-        GDK_SELECTION_PRIMARY,   /* VD_AGENT_CLIPBOARD_SELECTION_PRIMARY */
-    };
-#endif
+}
+
+VDAgentClipboards *vdagent_clipboards_new(struct vdagent_x11 *x11)
+{
+    VDAgentClipboards *self = g_object_new(VDAGENT_TYPE_CLIPBOARDS, NULL);
 
-    VDAgentClipboards *c;
-    c = g_new0(VDAgentClipboards, 1);
 #ifndef WITH_GTK
-    c->x11 = x11;
+    self->x11 = x11;
 #else
-    c->conn = conn;
+    guint sel_id;
 
     for (sel_id = 0; sel_id < SELECTION_COUNT; sel_id++) {
         GtkClipboard *clipboard = gtk_clipboard_get(sel_atom[sel_id]);
-        c->selections[sel_id].clipboard = clipboard;
-        /* enables the use of sel_id_from_clipboard(clipboard) macro */
-        g_object_set_data(G_OBJECT(clipboard), "vdagent-selection-id",
-                          GUINT_TO_POINTER(sel_id));
+        self->selections[sel_id].clipboard = clipboard;
         g_signal_connect(G_OBJECT(clipboard), "owner-change",
-                         G_CALLBACK(clipboard_owner_change_cb), c);
+                         G_CALLBACK(clipboard_owner_change_cb), self);
     }
 #endif
 
-    return c;
+    return self;
 }
 
-void vdagent_clipboards_finalize(VDAgentClipboards *c, gboolean conn_alive)
+void
+vdagent_clipboards_set_conn(VDAgentClipboards *self, UdscsConnection *conn)
+{
+    self->conn = conn;
+}
+
+static void vdagent_clipboards_dispose(GObject *obj)
 {
 #ifdef WITH_GTK
+    VDAgentClipboards *self = VDAGENT_CLIPBOARDS(obj);
     guint sel_id;
+
     for (sel_id = 0; sel_id < SELECTION_COUNT; sel_id++)
-        g_signal_handlers_disconnect_by_func(c->selections[sel_id].clipboard,
-            G_CALLBACK(clipboard_owner_change_cb), c);
+        g_signal_handlers_disconnect_by_func(self->selections[sel_id].clipboard,
+            G_CALLBACK(clipboard_owner_change_cb), self);
 
-    if (conn_alive == FALSE)
-        c->conn = NULL;
-    vdagent_clipboards_release_all(c);
+    if (self->conn)
+        vdagent_clipboards_release_all(self);
 #endif
+}
 
-    g_free(c);
+static void
+vdagent_clipboards_class_init(VDAgentClipboardsClass *klass)
+{
+    GObjectClass *oclass = G_OBJECT_CLASS(klass);
+
+    oclass->dispose = vdagent_clipboards_dispose;
 }
