@@ -529,14 +529,14 @@ void vdagent_x11_randr_handle_root_size_change(struct vdagent_x11 *x11,
 }
 
 int vdagent_x11_randr_handle_event(struct vdagent_x11 *x11,
-    XEvent event)
+    const XEvent *event)
 {
     int handled = TRUE;
 
-    switch (event.type - x11->xrandr_event_base) {
+    switch (event->type - x11->xrandr_event_base) {
         case RRScreenChangeNotify: {
-            XRRScreenChangeNotifyEvent *sce =
-                (XRRScreenChangeNotifyEvent *) &event;
+            const XRRScreenChangeNotifyEvent *sce =
+                (const XRRScreenChangeNotifyEvent *) event;
             vdagent_x11_randr_handle_root_size_change(x11, 0,
                 sce->width, sce->height);
             break;
@@ -555,16 +555,6 @@ int vdagent_x11_randr_handle_event(struct vdagent_x11 *x11,
     return handled;
 }
 
-static int min_int(int x, int y)
-{
-    return x > y ? y : x;
-}
-
-static int max_int(int x, int y)
-{
-    return x > y ? x : y;
-}
-
 static int constrain_to_range(int low, int *val, int high)
 {
     if (low <= *val && *val <= high) {
@@ -621,20 +611,24 @@ static void zero_base_monitors(struct vdagent_x11 *x11,
                                int *width, int *height)
 {
     int i, min_x = INT_MAX, min_y = INT_MAX, max_x = INT_MIN, max_y = INT_MIN;
-    int *mon_height, *mon_width;
 
     for (i = 0; i < mon_config->num_of_monitors; i++) {
-        if (!monitor_enabled(&mon_config->monitors[i]))
+        int mon_height, mon_width;
+
+        if (!monitor_enabled(&mon_config->monitors[i])) {
             continue;
+        }
         mon_config->monitors[i].x &= ~7;
         mon_config->monitors[i].width &= ~7;
-        mon_width = (int *)&mon_config->monitors[i].width;
-        mon_height = (int *)&mon_config->monitors[i].height;
-        constrain_to_screen(x11, mon_width, mon_height);
-        min_x = min_int(mon_config->monitors[i].x, min_x);
-        min_y = min_int(mon_config->monitors[i].y, min_y);
-        max_x = max_int(mon_config->monitors[i].x + *mon_width, max_x);
-        max_y = max_int(mon_config->monitors[i].y + *mon_height, max_y);
+        mon_width = mon_config->monitors[i].width;
+        mon_height = mon_config->monitors[i].height;
+        constrain_to_screen(x11, &mon_width, &mon_height);
+        min_x = MIN(mon_config->monitors[i].x, min_x);
+        min_y = MIN(mon_config->monitors[i].y, min_y);
+        max_x = MAX(mon_config->monitors[i].x + mon_width, max_x);
+        max_y = MAX(mon_config->monitors[i].y + mon_height, max_y);
+        mon_config->monitors[i].width = mon_width;
+        mon_config->monitors[i].height = mon_height;
     }
     if (min_x != 0 || min_y != 0) {
         syslog(LOG_ERR, "%s: agent config %d,%d rooted, adjusting to 0,0.",
