@@ -59,12 +59,12 @@ int vdagent_x11_caught_error;
 
 #ifndef WITH_GTK
 static void vdagent_x11_handle_selection_notify(struct vdagent_x11 *x11,
-                                                XEvent *event, int incr);
+                                                const XEvent *event, int incr);
 static void vdagent_x11_handle_selection_request(struct vdagent_x11 *x11);
 static void vdagent_x11_handle_targets_notify(struct vdagent_x11 *x11,
-                                              XEvent *event);
+                                              const XEvent *event);
 static void vdagent_x11_handle_property_delete_notify(struct vdagent_x11 *x11,
-                                                      XEvent *del_event);
+                                                      const XEvent *del_event);
 static void vdagent_x11_send_selection_notify(struct vdagent_x11 *x11,
                 Atom prop, struct vdagent_x11_selection_request *request);
 static void vdagent_x11_set_clipboard_owner(struct vdagent_x11 *x11,
@@ -196,7 +196,7 @@ static gchar *vdagent_x11_get_wm_name(struct vdagent_x11 *x11)
 #endif
 }
 
-struct vdagent_x11 *vdagent_x11_create(struct udscs_connection *vdagentd,
+struct vdagent_x11 *vdagent_x11_create(UdscsConnection *vdagentd,
     int debug, int sync)
 {
     struct vdagent_x11 *x11;
@@ -392,7 +392,7 @@ static void vdagent_x11_set_clipboard_owner(struct vdagent_x11 *x11,
                 once = 0;
             }
             vdagent_x11_send_selection_notify(x11, None, curr_sel);
-            if (curr_sel == x11->selection_req) {
+            if (prev_sel == NULL) {
                 x11->selection_req = next_sel;
                 free(x11->selection_req_data);
                 x11->selection_req_data = NULL;
@@ -400,7 +400,6 @@ static void vdagent_x11_set_clipboard_owner(struct vdagent_x11 *x11,
                 x11->selection_req_data_size = 0;
                 x11->selection_req_atom = None;
             } else {
-                // coverity[var_deref_op] if it is not the first there's a previous
                 prev_sel->next = next_sel;
             }
             free(curr_sel);
@@ -424,12 +423,11 @@ static void vdagent_x11_set_clipboard_owner(struct vdagent_x11 *x11,
             if (x11->vdagentd)
                 udscs_write(x11->vdagentd, VDAGENTD_CLIPBOARD_DATA, selection,
                             VD_AGENT_CLIPBOARD_NONE, NULL, 0);
-            if (curr_conv == x11->conversion_req) {
+            if (prev_conv == NULL) {
                 x11->conversion_req = next_conv;
                 x11->clipboard_data_size = 0;
                 x11->expect_property_notify = 0;
             } else {
-                // coverity[var_deref_op] if it is not the first there's a previous
                 prev_conv->next = next_conv;
             }
             free(curr_conv);
@@ -465,12 +463,13 @@ static int vdagent_x11_get_clipboard_atom(struct vdagent_x11 *x11, uint8_t selec
 }
 
 static int vdagent_x11_get_clipboard_selection(struct vdagent_x11 *x11,
-    XEvent *event, uint8_t *selection)
+    const XEvent *event, uint8_t *selection)
 {
     Atom atom;
 
     if (event->type == x11->xfixes_event_base) {
-        XFixesSelectionNotifyEvent *xfev = (XFixesSelectionNotifyEvent *)event;
+        const XFixesSelectionNotifyEvent *xfev =
+            (const XFixesSelectionNotifyEvent *)event;
         atom = xfev->selection;
     } else if (event->type == SelectionNotify) {
         atom = event->xselection.selection;
@@ -494,23 +493,23 @@ static int vdagent_x11_get_clipboard_selection(struct vdagent_x11 *x11,
 }
 #endif
 
-static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
+static void vdagent_x11_handle_event(struct vdagent_x11 *x11, const XEvent *event)
 {
     int i, handled = 0;
 #ifndef WITH_GTK
     uint8_t selection;
 
-    if (event.type == x11->xfixes_event_base) {
+    if (event->type == x11->xfixes_event_base) {
         union {
             XEvent ev;
             XFixesSelectionNotifyEvent xfev;
         } ev;
 
-        if (vdagent_x11_get_clipboard_selection(x11, &event, &selection)) {
+        if (vdagent_x11_get_clipboard_selection(x11, event, &selection)) {
             return;
         }
 
-        ev.ev = event;
+        ev.ev = *event;
         switch (ev.xfev.subtype) {
         case XFixesSetSelectionOwnerNotify:
             break;
@@ -521,7 +520,7 @@ static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
             break;
         default:
             VSELPRINTF("unexpected xfix event subtype %d window %d",
-                       (int)ev.xfev.subtype, (int)event.xany.window);
+                       (int)ev.xfev.subtype, (int)event->xany.window);
             return;
         }
         VSELPRINTF("New selection owner: %u", (unsigned int)ev.xfev.owner);
@@ -530,11 +529,10 @@ static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
         if (ev.xfev.owner == x11->selection_window)
             return;
 
-        /* If the clipboard owner is changed we no longer own it */
-        vdagent_x11_set_clipboard_owner(x11, selection, owner_none);
-
-        if (ev.xfev.owner == None)
+        if (ev.xfev.owner == None) {
+            vdagent_x11_set_clipboard_owner(x11, selection, owner_none);
             return;
+        }
 
         /* Request the supported targets from the new owner */
         XConvertSelection(x11->display, ev.xfev.selection, x11->targets_atom,
@@ -548,17 +546,17 @@ static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
     if (vdagent_x11_randr_handle_event(x11, event))
         return;
 
-    switch (event.type) {
+    switch (event->type) {
     case ConfigureNotify:
         for (i = 0; i < x11->screen_count; i++)
-            if (event.xconfigure.window == x11->root_window[i])
+            if (event->xconfigure.window == x11->root_window[i])
                 break;
         if (i == x11->screen_count)
             break;
 
         handled = 1;
         vdagent_x11_randr_handle_root_size_change(x11, i,
-                event.xconfigure.width, event.xconfigure.height);
+                event->xconfigure.width, event->xconfigure.height);
         break;
     case MappingNotify:
         /* These are uninteresting */
@@ -566,21 +564,21 @@ static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
         break;
 #ifndef WITH_GTK
     case SelectionNotify:
-        if (event.xselection.target == x11->targets_atom)
-            vdagent_x11_handle_targets_notify(x11, &event);
+        if (event->xselection.target == x11->targets_atom)
+            vdagent_x11_handle_targets_notify(x11, event);
         else
-            vdagent_x11_handle_selection_notify(x11, &event, 0);
+            vdagent_x11_handle_selection_notify(x11, event, 0);
 
         handled = 1;
         break;
     case PropertyNotify:
         if (x11->expect_property_notify &&
-                                event.xproperty.state == PropertyNewValue) {
-            vdagent_x11_handle_selection_notify(x11, &event, 1);
+                                event->xproperty.state == PropertyNewValue) {
+            vdagent_x11_handle_selection_notify(x11, event, 1);
         }
         if (x11->selection_req_data &&
-                                 event.xproperty.state == PropertyDelete) {
-            vdagent_x11_handle_property_delete_notify(x11, &event);
+                                 event->xproperty.state == PropertyDelete) {
+            vdagent_x11_handle_property_delete_notify(x11, event);
         }
         /* Always mark as handled, since we cannot unselect input for property
            notifications once we are done with handling the incr transfer. */
@@ -594,7 +592,7 @@ static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
     case SelectionRequest: {
         struct vdagent_x11_selection_request *req, *new_req;
 
-        if (vdagent_x11_get_clipboard_selection(x11, &event, &selection)) {
+        if (vdagent_x11_get_clipboard_selection(x11, event, &selection)) {
             return;
         }
 
@@ -606,7 +604,7 @@ static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
 
         handled = 1;
 
-        new_req->event = event;
+        new_req->event = *event;
         new_req->selection = selection;
         new_req->next = NULL;
 
@@ -628,7 +626,7 @@ static void vdagent_x11_handle_event(struct vdagent_x11 *x11, XEvent event)
     }
     if (!handled && x11->debug)
         syslog(LOG_DEBUG, "unhandled x11 event, type %d, window %d",
-               (int)event.type, (int)event.xany.window);
+               (int)event->type, (int)event->xany.window);
 }
 
 void vdagent_x11_do_read(struct vdagent_x11 *x11)
@@ -637,7 +635,7 @@ void vdagent_x11_do_read(struct vdagent_x11 *x11)
 
     while (XPending(x11->display)) {
         XNextEvent(x11->display, &event);
-        vdagent_x11_handle_event(x11, event);
+        vdagent_x11_handle_event(x11, &event);
     }
 }
 
@@ -650,7 +648,7 @@ static const char *vdagent_x11_get_atom_name(struct vdagent_x11 *x11, Atom a)
     return XGetAtomName(x11->display, a);
 }
 
-static int vdagent_x11_get_selection(struct vdagent_x11 *x11, XEvent *event,
+static int vdagent_x11_get_selection(struct vdagent_x11 *x11, const XEvent *event,
     uint8_t selection, Atom type, Atom prop, int format,
     unsigned char **data_ret, int incr)
 {
@@ -842,7 +840,7 @@ static void vdagent_x11_handle_conversion_request(struct vdagent_x11 *x11)
 }
 
 static void vdagent_x11_handle_selection_notify(struct vdagent_x11 *x11,
-                                                XEvent *event, int incr)
+                                                const XEvent *event, int incr)
 {
     int len = 0;
     unsigned char *data = NULL;
@@ -927,7 +925,7 @@ static void vdagent_x11_print_targets(struct vdagent_x11 *x11,
 }
 
 static void vdagent_x11_handle_targets_notify(struct vdagent_x11 *x11,
-                                              XEvent *event)
+                                              const XEvent *event)
 {
     int i, len;
     Atom atom, *atoms = NULL;
@@ -1026,7 +1024,7 @@ static void vdagent_x11_send_selection_notify(struct vdagent_x11 *x11,
 }
 
 static void vdagent_x11_send_targets(struct vdagent_x11 *x11,
-    uint8_t selection, XEvent *event)
+    uint8_t selection, const XEvent *event)
 {
     Atom prop, targets[256] = { x11->targets_atom, };
     int i, j, k, target_count = 1;
@@ -1123,7 +1121,7 @@ static void vdagent_x11_handle_selection_request(struct vdagent_x11 *x11)
 }
 
 static void vdagent_x11_handle_property_delete_notify(struct vdagent_x11 *x11,
-                                                      XEvent *del_event)
+                                                      const XEvent *del_event)
 {
     XEvent *sel_event;
     int len;
@@ -1370,7 +1368,7 @@ void vdagent_x11_clipboard_release(struct vdagent_x11 *x11, uint8_t selection)
     XSync(x11->display, False);
     while (XCheckTypedEvent(x11->display, x11->xfixes_event_base,
                             &event))
-        vdagent_x11_handle_event(x11, event);
+        vdagent_x11_handle_event(x11, &event);
 
     /* Note no need to do a set_clipboard_owner(owner_none) here, as that is
        already done by processing the XFixesSetSelectionOwnerNotify event. */
