@@ -23,7 +23,7 @@
 #ifndef __VDAGENTD_PROTO_H
 #define __VDAGENTD_PROTO_H
 
-#define VDAGENTD_SOCKET "/var/run/spice-vdagentd/spice-vdagent-sock"
+#define VDAGENTD_SOCKET "/run/spice-vdagentd/spice-vdagent-sock"
 
 #define DEFAULT_VIRTIO_PORT_PATH "/dev/virtio-ports/com.redhat.spice.0"
 
@@ -44,6 +44,7 @@ enum {
     VDAGENTD_FILE_XFER_DATA,
     VDAGENTD_FILE_XFER_DISABLE,
     VDAGENTD_CLIENT_DISCONNECTED,  /* daemon -> client */
+    VDAGENTD_GRAPHICS_DEVICE_INFO,  /* daemon -> client */
     VDAGENTD_NO_MESSAGES /* Must always be last */
 };
 
@@ -52,6 +53,7 @@ struct vdagentd_guest_xorg_resolution {
     int height;
     int x;
     int y;
+    int display_id;
 };
 
 #endif
